@@ -4,14 +4,14 @@
 abs() {
 	local _n
 
-	_n="$1"; : "{$_n:?}"
+	_n="$1"; : "${_n:?}"
 
 	: "${_n:?}"
 
-	if [ "$_n" -lt 0 ]; then
+	if [ "${_n}" -lt 0 ]; then
 		echo "$((- _n))"
 	else
-		echo "$_n"
+		echo "${_n}"
 	fi
 }
 
@@ -30,7 +30,7 @@ build_date() {
 	done
 	: "${_builddir:?}"
 
-	step_eval 1 "$(step_path "$_builddir")"
+	step_eval 1 "$(step_path "${_builddir}")"
 	step_value time
 }
 
@@ -43,7 +43,7 @@ build_id() {
 
 	_d="$(date '+%Y-%m-%d')"
 	_c="$(find "$1" -type d -name "${_d}*" | wc -l)"
-	printf '%s.%d\n' "$_d" "$((_c + 1))"
+	printf '%s.%d\n' "${_d}" "$((_c + 1))"
 }
 
 # build_init build-dir
@@ -54,12 +54,12 @@ build_init() {
 	local _steps
 
 	_builddir="$1"; : "${_builddir:?}"
-	_steps="$(step_path "$_builddir")"
+	_steps="$(step_path "${_builddir}")"
 
-	[ -d "$_builddir" ] || mkdir "$_builddir"
+	[ -d "${_builddir}" ] || mkdir "${_builddir}"
 	[ -d "${_builddir}/tmp" ] || mkdir "${_builddir}/tmp"
 	[ -e "${_builddir}/robsd.log" ] || : >"${_builddir}/robsd.log"
-	[ -e "$_steps" ] || : >"$_steps"
+	[ -e "${_steps}" ] || : >"${_steps}"
 	return 0
 }
 
@@ -76,7 +76,7 @@ check_perf() {
 	esac
 
 	_perf="$(sysctl -n hw.setperf 2>/dev/null)"
-	if [ -z "$_perf" ] || [ "$_perf" -eq 100 ]; then
+	if [ -z "${_perf}" ] || [ "${_perf}" -eq 100 ]; then
 		return 0
 	fi
 
@@ -92,7 +92,7 @@ cleandir() {
 	local _d
 
 	for _d; do
-                find "$_d" -mindepth 1 -maxdepth 1 -print0 | xargs -0r rm -r
+                find "${_d}" -mindepth 1 -maxdepth 1 -print0 | xargs -0r rm -r
 	done
 }
 
@@ -107,7 +107,7 @@ config_load() {
 	: "${BUILDDIR:=}"
 	: "${DETACH:=1}"
 	: "${EXECDIR:=/usr/local/libexec/robsd}"; export EXECDIR
-	PATH="${PATH}:/usr/X11R6/bin"; export PATH
+	: "${ROBSDCLEAN:=/usr/local/sbin/robsd-clean}"
 	: "${ROBSDCONFIG:=${EXECDIR}/robsd-config}"
 	: "${ROBSDEXEC:=${EXECDIR}/robsd-exec}"
 	: "${ROBSDHOOK:=${EXECDIR}/robsd-hook}"
@@ -120,15 +120,15 @@ config_load() {
 	_tmp="$(mktemp -t robsd.XXXXXX)"
 	{
 		cat
-	} | "$ROBSDCONFIG" -m "$_MODE" ${ROBSDCONF:+"-C${ROBSDCONF}"} "$@" - \
-		>"$_tmp" || _err="$?"
-	[ "$_err" -eq 0 ] && eval "$(<"$_tmp")"
-	rm "$_tmp"
-	[ "$_err" -eq 0 ] || return "$_err"
+	} | "${ROBSDCONFIG}" -m "${_MODE}" ${ROBSDCONF:+"-C${ROBSDCONF}"} "$@" - \
+		>"${_tmp}" || _err="$?"
+	[ "${_err}" -eq 0 ] && eval "$(<"${_tmp}")"
+	rm "${_tmp}"
+	[ "${_err}" -eq 0 ] || return "${_err}"
 
-	case "$_MODE" in
+	case "${_MODE}" in
 	robsd|robsd-cross)
-		MAKEFLAGS="-j$(sysctl -n hw.ncpuonline)"; export MAKEFLAGS
+		MAKEFLAGS="-j$(config_value ncpu)"; export MAKEFLAGS
 		;;
 	robsd-ports)
 		ports_config_load
@@ -136,6 +136,8 @@ config_load() {
 	robsd-regress)
 		regress_config_load
 		;;
+	canvas)
+		;;
 	*)
 		return 1
 		;;
@@ -152,7 +154,7 @@ config_value()
 	_var="$1"; : "${_var:?}"
 
 	echo "\${${_var}}" |
-	"$ROBSDCONFIG" -m "$_MODE" ${ROBSDCONF:+"-C${ROBSDCONF}"} -
+	"${ROBSDCONFIG}" -m "${_MODE}" ${ROBSDCONF:+"-C${ROBSDCONF}"} -
 }
 
 # cvs_changelog -t tmp-dir
@@ -177,8 +179,8 @@ cvs_changelog() {
 	${_tmpdir}/cvs-xenocara-ci.log
 	EOF
 	do
-		[ -s "$_f" ] || continue
-		cat "$_f"; echo
+		[ -s "${_f}" ] || continue
+		cat "${_f}"; echo
 	done
 }
 
@@ -192,10 +194,10 @@ cvs_field() {
 	_field="$1"; : "${_field:?}"
 	_line="$2"; : "${_line:?}"
 
-	echo "$_line" | grep -q -F "$_field" || return 1
+	echo "${_line}" | grep -q -F "${_field}" || return 1
 
 	_line="${_line##*"${_field}": }"; _line="${_line%%;*}"
-	echo "$_line"
+	echo "${_line}"
 }
 
 # cvs_date -b build-dir -s steps
@@ -218,16 +220,16 @@ cvs_date() {
 	: "${_builddir:?}"
 	: "${_steps:?}"
 
-	step_eval -n cvs "$_steps"
+	step_eval -n cvs "${_steps}"
 	step_skip && return 1
 
 	# Try to find the date of the last revision in the log, i.e. the first
 	# entry written by cvs_log(). If nothing was updated, use the step
 	# execution date of the cvs step as a fallback.
 	_log="${_builddir}/$(step_value log 2>/dev/null || :)"
-	_date="$(grep -m 1 '^Date:' "$_log" | sed -e 's/^[^:]*: *//' || :)"
-	if [ -n "$_date" ]; then
-		date -j -f '%Y/%m/%d %H:%M:%S' '+%s' "$_date"
+	_date="$(grep -m 1 '^Date:' "${_log}" | sed -e 's/^[^:]*: *//' || :)"
+	if [ -n "${_date}" ]; then
+		date -j -f '%Y/%m/%d %H:%M:%S' '+%s' "${_date}"
 	else
 		step_value time
 	fi
@@ -266,44 +268,44 @@ cvs_log() {
 	: "${_host:?}"
 	: "${_user:?}"
 
-	[ -d "$_tmp" ] && rm -r "$_tmp"
-	mkdir -p "$_tmp"
+	[ -d "${_tmp}" ] && rm -r "${_tmp}"
+	mkdir -p "${_tmp}"
 
 	# Use the date from latest revision from the previous release.
 	for _prev in $(prev_release -B); do
-		_date="$(cvs_date -b "$_prev" -s "$(step_path "$_prev")")" && break
+		_date="$(cvs_date -b "${_prev}" -s "$(step_path "${_prev}")")" && break
 	done
-	if [ -z "$_date" ]; then
+	if [ -z "${_date}" ]; then
 		echo "cvs_log: previous date not found" 1>&2
 		return 0
 	fi
-	_date="$(date -r "$_date" '+%F %T')"
+	_date="$(date -r "${_date}" '+%F %T')"
 
 	grep '^[MPU]\>' |
 	cut -d ' ' -f 2 |
-	unpriv "$_user" "cd ${_repo} && xargs cvs -q -d ${_host} log -N -l -d '>${_date}'" |
+	unpriv "${_user}" "cd ${_repo} && xargs cvs -q -d ${_host} log -N -l -d '>${_date}'" |
 	tee "${_tmp}/cvs.log" |
 	while read -r _line; do
-		case "$_line" in
+		case "${_line}" in
 		Working\ file:*)
 			_path="${_line#*: }"
 			;;
 		date:*)
-			_date="$(cvs_field date "$_line")"
-			_cid="$(cvs_field commitid "$_line")" || continue
+			_date="$(cvs_field date "${_line}")"
+			_cid="$(cvs_field commitid "${_line}")" || continue
 			if ! [ -d "${_tmp}/${_cid}" ]; then
 				mkdir "${_tmp}/${_cid}"
-				cvs_field author "$_line" >"${_tmp}/${_cid}/author"
+				cvs_field author "${_line}" >"${_tmp}/${_cid}/author"
 				_message=1
 			fi
-			echo "$_date" >>"${_tmp}/${_cid}/date"
+			echo "${_date}" >>"${_tmp}/${_cid}/date"
 			echo "${_indent}${_path}" >>"${_tmp}/${_cid}/files"
 			;;
 		-[-]*|=[=]*)
 			_message=0
 			;;
 		*)
-			if [ "$_message" -eq 1 ]; then
+			if [ "${_message}" -eq 1 ]; then
 				echo "${_indent}${_line}" >>"${_tmp}/${_cid}/message"
 			fi
 			;;
@@ -311,9 +313,9 @@ cvs_log() {
 	done
 
 	# Sort each commit using the date file.
-	find "$_tmp" -type f -name date |
+	find "${_tmp}" -type f -name date |
 	while read -r _p; do
-		_date="$(sort -nr "$_p" | head -1)"
+		_date="$(sort -nr "${_p}" | head -1)"
 		echo "${_date} ${_p%/*}"
 	done |
 	sort -nr |
@@ -358,27 +360,27 @@ diff_apply() (
 	: "${_user:?}"
 	: "${_diff:?}"
 
-	_root="$(diff_root -d "$_dir" "$_diff")"
-	cd "$_root"
+	_root="$(diff_root -d "${_dir}" "${_diff}")"
+	cd "${_root}"
 
 	# Try to revert the diff if dry run fails.
-	if ! unpriv "$_user" "exec patch -C -Efs" <"$_diff" >/dev/null; then
-		unpriv "$_user" "exec patch -R -Efs" <"$_diff"
+	if ! unpriv "${_user}" "exec patch -C -Efs" <"${_diff}" >/dev/null; then
+		unpriv "${_user}" "exec patch -R -Efs" <"${_diff}"
 	fi
 	# Use the strip argument in order to cope with files in newly created
 	# directories since they would otherwise end up in the current working
 	# directory. However, we could operate on a Git diff in which prefixes
 	# must be stripped.
 	for _strip in 0 1; do
-		if unpriv "$_user" "exec patch -Efs -p ${_strip}" \
-		   <"$_diff" >"$_tmp" 2>&1; then
+		if unpriv "${_user}" "exec patch -Efs -p ${_strip}" \
+		   <"${_diff}" >"${_tmp}" 2>&1; then
 			break
 		fi
 	done
-	[ -s "$_tmp" ] && _err=1
-	cat "$_tmp"
-	rm -f "$_tmp"
-	return "$_err"
+	[ -s "${_tmp}" ] && _err=1
+	cat "${_tmp}"
+	rm -f "${_tmp}"
+	return "${_err}"
 )
 
 # diff_clean dir
@@ -390,11 +392,11 @@ diff_clean() {
 
 	_dir="$1"; : "${_dir:?}"
 
-	find "$_dir" -type f \( \
+	find "${_dir}" -type f \( \
 		-name '*.orig' -o -name '*.rej' -o -name '.#*' \) |
 	while read -r _path; do
 		if ! grep -sq "^/${_path##*/}/" "${_path%/*}/CVS/Entries"; then
-			echo "$_path"
+			echo "${_path}"
 		fi
 	done |
 	xargs -r rm
@@ -426,16 +428,16 @@ diff_copy() {
 	for _src; do
 		_dst="${_base}.${_i}"
 
-		_r="$(diff_root -d "$_root" "$_src")"
+		_r="$(diff_root -d "${_root}" "${_src}")"
 		info "using diff ${_src} rooted in ${_r}"
 
 		{
-			if ! head -1 "$_src" | grep -q '^#'; then
-				printf '# %s\n\n' "$_src"
+			if ! head -1 "${_src}" | grep -q '^#'; then
+				printf '# %s\n\n' "${_src}"
 			fi
-			cat "$_src"
-		} >"$_dst"
-		chmod 644 "$_dst"
+			cat "${_src}"
+		} >"${_dst}"
+		chmod 644 "${_dst}"
 
 		_i=$((_i + 1))
 	done
@@ -451,7 +453,7 @@ diff_list() {
 	_builddir="$1" ; : "${_builddir:?}"
 	_prefix="$2" ; : "${_prefix:?}"
 
-	find "$_builddir" -maxdepth 1 -type f -name "${_prefix}.*" | sort
+	find "${_builddir}" -maxdepth 1 -type f -name "${_prefix}.*" | sort
 }
 
 # diff_revert -d dir -t tmp-dir -u user diff
@@ -480,30 +482,30 @@ diff_revert() (
 	: "${_user:?}"
 	: "${_diff:?}"
 
-	_root="$(diff_root -d "$_dir" "$_diff")"
-	cd "$_root"
+	_root="$(diff_root -d "${_dir}" "${_diff}")"
+	cd "${_root}"
 
-	if unpriv "$_user" "exec patch -CR -Efs" <"$_diff" >/dev/null 2>&1; then
+	if unpriv "${_user}" "exec patch -CR -Efs" <"${_diff}" >/dev/null 2>&1; then
 		info "reverting diff ${_diff}"
-		unpriv "$_user" "exec patch -R -Ef" <"$_diff" >"$_revert"
+		unpriv "${_user}" "exec patch -R -Ef" <"${_diff}" >"${_revert}"
 	else
 		info "diff already reverted ${_diff}"
 	fi
-	if [ -e "$_revert" ]; then
-		diff_clean "$_dir"
+	if [ -e "${_revert}" ]; then
+		diff_clean "${_dir}"
 
 		# Remove empty directories.
-		sed -n -e 's/^Removing \([^[:space:]]*\) (empty .*/\1/p' "$_revert" |
+		sed -n -e 's/^Removing \([^[:space:]]*\) (empty .*/\1/p' "${_revert}" |
 		xargs -r -L 1 dirname |
 		sort |
 		uniq |
 		while read -r _p; do
-			isempty "$_p" || continue
+			isempty "${_p}" || continue
 			info "removing empty directory ${_root}/${_p}"
-			rmdir "$_p"
+			rmdir "${_p}"
 		done
 	fi
-	rm -f "$_revert"
+	rm -f "${_revert}"
 )
 
 # diff_root -d directory diff
@@ -533,18 +535,18 @@ diff_root() {
 	xargs |
 	while read -r _file _path; do
 		_p="${_path%/"${_file}"}"
-		while [ -n "$_p" ]; do
+		while [ -n "${_p}" ]; do
 			if [ -e "${_root}${_p}" ]; then
 				echo "${_root}${_p}"
 				_err=0
 				break
 			fi
 
-			_p="$(path_strip "$_p")"
+			_p="$(path_strip "${_p}")"
 		done
 
-		return "$_err"
-	done || echo "$_root"
+		return "${_err}"
+	done || echo "${_root}"
 
 	return 0
 }
@@ -563,14 +565,14 @@ duration_prev() {
 
 	prev_release -B |
 	while read -r _prev; do
-		step_eval -n "$_step" "$(step_path "$_prev")" 2>/dev/null || continue
+		step_eval -n "${_step}" "$(step_path "${_prev}")" 2>/dev/null || continue
 		step_skip && continue
 
 		_exit="$(step_value exit 2>/dev/null || echo 1)"
-		[ "$_exit" -eq 0 ] || continue
+		[ "${_exit}" -eq 0 ] || continue
 
 		_duration="$(step_value duration 2>/dev/null)" || continue
-		echo "$_duration"
+		echo "${_duration}"
 		return 1
 	done || return 0
 
@@ -595,16 +597,16 @@ duration_total() {
 	done
 	: "${_steps:?}"
 
-	case "$_MODE" in
+	case "${_MODE}" in
 	robsd-regress)
-		regress_duration_total -s "$_steps"
+		regress_duration_total -s "${_steps}"
 		return 0
 		;;
 	*)
 		;;
 	esac
 
-	while step_eval "$_i" "$_steps" 2>/dev/null; do
+	while step_eval "${_i}" "${_steps}" 2>/dev/null; do
 		_i=$((_i + 1))
 
 		step_skip && continue
@@ -617,7 +619,7 @@ duration_total() {
 		_tot=$((_tot + _d))
 	done
 
-	echo "$_tot"
+	echo "${_tot}"
 }
 
 # fatal message ...
@@ -638,7 +640,7 @@ info() {
 		# Not fully detached yet, write all entries to robsd.log.
 		_log="${BUILDDIR}/robsd.log"
 	fi
-	echo "${_PROG}: ${*}" | tee -a "$_log"
+	echo "${_PROG}: ${*}" | tee -a "${_log}"
 }
 
 # isempty path
@@ -648,7 +650,7 @@ isempty() {
 	local _path
 
 	_path="$1"; : "${_path:?}"
-	! find "$_path" -empty | cmp -s - /dev/null
+	! find "${_path}" -empty | cmp -s - /dev/null
 }
 
 # jobs_count job ...
@@ -674,12 +676,12 @@ lock_acquire() {
 	# We could already be owning the lock if the previous run was aborted
 	# prematurely.
 	_owner="$(cat "${_rootdir}/.running" 2>/dev/null || :)"
-	if [ -n "$_owner" ] && [ "$_owner" != "$_builddir" ]; then
+	if [ -n "${_owner}" ] && [ "${_owner}" != "${_builddir}" ]; then
 		info "${_owner}: lock already acquired"
 		return 1
 	fi
 
-	echo "$_builddir" >"${_rootdir}/.running"
+	echo "${_builddir}" >"${_rootdir}/.running"
 }
 
 # lock_alive root-dir build-dir
@@ -693,7 +695,7 @@ lock_alive() {
 	_rootdir="$1"; : "${_rootdir:?}"
 	_builddir="$2"; : "${_builddir:?}"
 	touch "${_rootdir}/.running" 2>/dev/null || return 1
-	echo "$_builddir" | cmp -s - "${_rootdir}/.running"
+	echo "${_builddir}" | cmp -s - "${_rootdir}/.running"
 }
 
 # lock_release root-dir build-dir
@@ -706,7 +708,7 @@ lock_release() {
 	_rootdir="$1"; : "${_rootdir:?}"
 	_builddir="$2"; : "${_builddir:?}"
 
-	if echo "$_builddir" | cmp -s - "${_rootdir}/.running"; then
+	if echo "${_builddir}" | cmp -s - "${_rootdir}/.running"; then
 		chflags nouchg "${_rootdir}/.running"
 		rm -f "${_rootdir}/.running"
 	else
@@ -736,13 +738,13 @@ log_id() {
 	: "${_builddir:?}"
 	: "${_step:?}"
 
-	_name="$(echo "$_name" | tr '/' '-')"
-	_id="$(printf '%03d-%s.log' "$_step" "$_name")"
-	_dups="$(find "$_builddir" -name "${_id}*" | wc -l)"
-	if [ "$_dups" -gt 0 ]; then
-		printf '%s.%d' "$_id" "$_dups"
+	_name="$(echo "${_name}" | tr '/' '-')"
+	_id="$(printf '%03d-%s.log' "${_step}" "${_name}")"
+	_dups="$(find "${_builddir}" -name "${_id}*" | wc -l)"
+	if [ "${_dups}" -gt 0 ]; then
+		printf '%s.%d' "${_id}" "${_dups}"
 	else
-		echo "$_id"
+		echo "${_id}"
 	fi
 }
 
@@ -756,7 +758,7 @@ path_strip() {
 	_src="$1"; : "${_src:?}"
 
 	_dst="${_src#/}"
-	[ "${_dst#*/}" = "$_dst" ] && return 0
+	[ "${_dst#*/}" = "${_dst}" ] && return 0
 	echo "/${_dst#*/}"
 
 }
@@ -765,10 +767,10 @@ path_strip() {
 #
 # Get previous invocations.
 prev_release() {
-	"$ROBSDLS" -m "${_MODE}" ${ROBSDCONF:+"-C${ROBSDCONF}"} "$@"
+	"${ROBSDLS}" -m "${_MODE}" ${ROBSDCONF:+"-C${ROBSDCONF}"} "$@"
 }
 
-# purge dir count
+# purge [-d] dir count
 #
 # Keep the latest count number of release directories in dir.
 # The older ones will be moved to the keep-dir, preserving only the relevant
@@ -777,28 +779,45 @@ purge() {
 	local _attic
 	local _d
 	local _dir
+	local _dry=0
 	local _dst
 	local _n
 	local _tim
 
+	while [ $# -gt 0 ]; do
+		case "$1" in
+		-d)	_dry=1;;
+		*)	break;;
+		esac
+		shift
+	done
 	_dir="$1"; : "${_dir:?}"
 	_n="$2"; : "${_n:?}"
 
 	_attic="$(config_value keep-dir)"
 
+	# While not running, must compensate for current builddir not being
+	# excluded by robsd-ls.
+	if ! config_value builddir >/dev/null 2>&1; then
+		_n="$((_n + 1))"
+	fi
+
 	prev_release -B |
-	sort -nr |
-	tail -n "+$((_n + 1))" |
+	tail -n "+${_n}" |
 	while read -r _d; do
-		[ -d "$_attic" ] || mkdir "$_attic"
+		if [ "${_dry}" -eq 1 ]; then
+			echo "${_d}"
+			continue
+		fi
+
+		[ -d "${_attic}" ] || mkdir "${_attic}"
 
 		# Grab the modification time before removal of irrelevant files.
-		_tim="$(stat -f '%Sm' -t '%FT%T' "$_d")"
+		_tim="$(stat -f '%Sm' -t '%FT%T' "${_d}")"
 
 		rm -rf "${_d}/tmp"
-		find "$_d" -mindepth 1 -not \( \
+		find "${_d}" -mindepth 1 -not \( \
 			-name '*.diff.*' -o \
-			-name '01-env.log' -o \
 			-name 'comment' -o \
 			-name 'index.txt' -o \
 			-name 'report' -o \
@@ -810,10 +829,10 @@ purge() {
 		_dst="${_attic}/$(echo "${_d##*/}" | tr '-' '/')"
 		# Create leading YYYY/MM directories.
 		mkdir -p "${_dst%/*}"
-		cp -pr "$_d" "$_dst"
-		touch -d "$_tim" "$_dst"
-		rm -r "$_d"
-		echo "$_d"
+		cp -pr "${_d}" "${_dst}"
+		touch -d "${_tim}" "${_dst}"
+		rm -r "${_d}"
+		echo "${_d}"
 	done
 }
 
@@ -823,7 +842,6 @@ purge() {
 report() {
 	local _builddir
 	local _report
-	local _steps
 
 	while [ $# -gt 0 ]; do
 		case "$1" in
@@ -834,14 +852,38 @@ report() {
 	done
 	: "${_builddir:?}"
 
-	_report="${_builddir}/report"
-	_steps="$(step_path "$_builddir")"
+	_report="$(config_value report-path)"
 
-	# The steps file could be empty when a build fails to start due to
-	# another already running build.
-	[ -s "$_steps" ] || return 1
+	"${ROBSDREPORT}" -m "${_MODE}" ${ROBSDCONF:+-C ${ROBSDCONF}} "${_builddir}" >"${_report}"
+}
 
-	"$ROBSDREPORT" -m "$_MODE" ${ROBSDCONF:+-C ${ROBSDCONF}} "$_builddir" >"$_report"
+# report_receiver -b build-dir
+#
+# Get report mail receiver.
+report_receiver() {
+	local _builddir
+	local _steps
+
+	while [ $# -gt 0 ]; do
+		case "$1" in
+		-b)	shift; _builddir="${1}";;
+		*)	break;;
+		esac
+		shift
+	done
+	: "${_builddir:?}"
+
+	_steps="$(step_path "${_builddir}")"
+
+	case "${_MODE}" in
+	canvas)
+		step_eval 1 "${_steps}"
+		step_value user
+		;;
+	*)
+		echo root
+		;;
+	esac
 }
 
 # robsd -b build-dir -s step-id
@@ -853,7 +895,7 @@ robsd() {
 	local _jobs=""
 	local _name
 	local _ncpu
-	local _s
+	local _parallel
 	local _step
 	local _steps
 
@@ -868,72 +910,68 @@ robsd() {
 	: "${_builddir:?}"
 	: "${_step:?}"
 
-	_ncpu="$(sysctl -n hw.ncpuonline)"
-	_steps="$(step_path "$_builddir")"
+	_ncpu="$(config_value ncpu)"
+	_steps="$(step_path "${_builddir}")"
 
-	while :; do
-		_name="$(step_name "$_step")"
-		_s="$_step"
-		_step=$((_step + 1))
-
-		if step_eval -n "$_name" "$_steps" 2>/dev/null &&
+	steps -o "${_step}" | while read -r _step _name _parallel; do
+		if step_eval -n "${_name}" "${_steps}" 2>/dev/null &&
 		   step_skip; then
 			info "step ${_name} skipped"
 			continue
 		fi
 		info "step ${_name}"
 
-		if [ "$_name" = "end" ]; then
-			# The duration of the end step is the accumulated
-			# duration.
-			_d1="$(duration_total -s "$_steps")"
-			_d0="$(duration_prev "$_name" || :)"
-			if [ -n "$_d0" ]; then
-				_delta="$((_d1 - _d0))"
-			else
-				_delta=0
-			fi
-			step_write -t -s "$_s" -n "$_name" -e 0 \
-				-d "$_d1" -a "$_delta" "$_steps"
-			# The hook is invoked as late as possible in the exit
-			# trap handler.
-			return 0
-		fi
-
-		if step_parallel "$_name"; then
+		if [ -n "${_parallel}" ]; then
 			# Ensure the job queue is not full.
-			if [ "$(jobs_count "$_jobs")" -eq "$_ncpu" ]; then
+			if [ "$(jobs_count "${_jobs}")" -eq "${_ncpu}" ]; then
 				info "parallel wait $(jobs_count "${_jobs}")/${_ncpu}"
-				_jobs="$(echo "$_jobs" | xargs "$ROBSDWAIT" | xargs)"
+				_jobs="$(echo "${_jobs}" | xargs "${ROBSDWAIT}" | xargs)"
 			fi
 
 			# Execute job in parallel.
-			step_exec_job -b "$_builddir" -s "$_steps" \
-				-i "$_s" -n "$_name" &
+			step_exec_job -b "${_builddir}" -s "${_steps}" \
+				-i "${_step}" -n "${_name}" &
 			_jobs="${_jobs}${_jobs:+ }${!}"
 			info "parallel jobs $(jobs_count "${_jobs}")/${_ncpu}"
 		else
 			# Wait for all running jobs to finish.
-			if [ -n "$_jobs" ]; then
+			if [ -n "${_jobs}" ]; then
 				info "parallel barrier $(jobs_count "${_jobs}")/${_ncpu}"
-				echo "$_jobs" | xargs "$ROBSDWAIT" -a
+				echo "${_jobs}" | xargs "${ROBSDWAIT}" -a
 				_jobs=""
 			fi
 
+			if [ "${_name}" = "end" ]; then
+				# The duration of the end step is the
+				# accumulated duration.
+				_d1="$(duration_total -s "${_steps}")"
+				_d0="$(duration_prev "${_name}" || :)"
+				if [ -n "${_d0}" ]; then
+					_delta="$((_d1 - _d0))"
+				else
+					_delta=0
+				fi
+				step_write -t -s "${_step}" -n "${_name}" -e 0 \
+					-d "${_d1}" -a "${_delta}" "${_steps}"
+				# The hook is invoked as late as possible in the
+				# exit trap handler.
+				return 0
+			fi
+
 			# Execute job synchronously.
-			step_exec_job -b "$_builddir" -s "$_steps" \
-				-i "$_s" -n "$_name"
+			step_exec_job -b "${_builddir}" -s "${_steps}" \
+				-i "${_step}" -n "${_name}"
 		fi
 
 		# Reboot in progress?
-		if [ "$_name" = "reboot" ] &&
+		if [ "${_name}" = "reboot" ] &&
 		   [ "$(config_value reboot)" -eq 1 ]; then
 			return 0
 		fi
 
 		# Does robsd-kill want us dead?
-		if ! lock_alive "$ROBSDDIR" "$_builddir"; then
-			[ -z "$_jobs" ] || echo "$_jobs" | xargs "$ROBSDWAIT" -a
+		if ! lock_alive "${ROBSDDIR}" "${_builddir}"; then
+			[ -z "${_jobs}" ] || echo "${_jobs}" | xargs "${ROBSDWAIT}" -a
 			return 1
 		fi
 	done
@@ -944,35 +982,18 @@ robsd() {
 # Invoke robsd hook.
 robsd_hook() {
 	# Ignore non-zero exit.
-	"$ROBSDHOOK" -m "$_MODE" -V ${ROBSDCONF:+"-C${ROBSDCONF}"} "$@" || :
+	"${ROBSDHOOK}" -m "${_MODE}" -V ${ROBSDCONF:+"-C${ROBSDCONF}"} "$@" || :
 }
 
 # setmode mode
-# setmode -p path
 #
-# Set the execution mode or infer it from the given program name.
+# Set the execution mode.
 setmode() {
 	local _mode=""
-	local _path=""
 
-	while [ $# -gt 0 ]; do
-		case "$1" in
-		-p)	shift; _path="$1";;
-		*)	break;;
-		esac
-		shift
-	done
+	_mode="$1"; : "${_mode:?}"
 
-	if [ -n "$_path" ]; then
-		case "${_path##*/}" in
-		robsd-ports*)	_mode="robsd-ports";;
-		robsd-regress*)	_mode="robsd-regress";;
-		*)		_mode="robsd";;
-		esac
-	else
-		_mode="$1"
-	fi
-	_MODE="$_mode"; export _MODE
+	_MODE="${_mode}"; export _MODE
 }
 
 # setprogname name
@@ -1017,7 +1038,7 @@ step_write() {
 
 	_user="$(logname)"
 
-	"$ROBSDSTEP" -W -f "$_file" -i "$_s" -- \
+	"${ROBSDSTEP}" -W -f "${_file}" -i "${_s}" -- \
 		"name=${_name}" \
 		"exit=${_exit}" \
 		"duration=${_duration}" \
@@ -1028,11 +1049,33 @@ step_write() {
 		"skip=${_skip}"
 }
 
+# has_steps file
+#
+# Returns zero if the given step file is not empty.
+has_steps() {
+	local _file
+	local _i=1
+
+	_file="$1"; : "${_file:?}"
+
+	while step_eval "${_i}" "${_file}" 2>/dev/null; do
+		_i="$((_i + 1))"
+
+		if [ "$(step_value skip)" -eq 1 ]; then
+			continue
+		else
+			return 0
+		fi
+	done
+
+	return 1
+}
+
 # step_eval offset file
 # step_eval -offset file
 # step_eval -n step-name file
 #
-# Read the given step using robsd-step into the _STEP array.
+# Read the given step using robsd-step into distinct variables for each field.
 step_eval() {
 	local _err=0
 	local _file
@@ -1052,60 +1095,72 @@ step_eval() {
 
 	_tmp="$(mktemp -t robsd.XXXXXX)"
 	{
-		printf 'set -A _STEP -- '
 		# shellcheck disable=SC2016
-		printf '"${step}" "${name}" "${exit}" "${duration}" "${log}" '
+		printf '_STEP_step=${step}\n'
 		# shellcheck disable=SC2016
-		printf '"${time}" "${user}" "${skip}" "${delta}"\n'
-	} | "$ROBSDSTEP" -R -f "$_file" "$_flag" "$_step" >"$_tmp" || _err="$?"
-	[ "$_err" -eq 0 ] && eval "$(<"$_tmp")"
-	rm "$_tmp"
-	return "$_err"
+		printf '_STEP_name=${name}\n'
+		# shellcheck disable=SC2016
+		printf '_STEP_exit=${exit}\n'
+		# shellcheck disable=SC2016
+		printf '_STEP_duration=${duration}\n'
+		# shellcheck disable=SC2016
+		printf '_STEP_log=${log}\n'
+		# shellcheck disable=SC2016
+		printf '_STEP_time=${time}\n'
+		# shellcheck disable=SC2016
+		printf '_STEP_user=${user}\n'
+		# shellcheck disable=SC2016
+		printf '_STEP_skip=${skip}\n'
+		# shellcheck disable=SC2016
+		printf '_STEP_delta=${delta}\n'
+	} | "${ROBSDSTEP}" -R -f "${_file}" "${_flag}" "${_step}" >"${_tmp}" || _err="$?"
+	[ "${_err}" -eq 0 ] && eval "$(<"${_tmp}")"
+	rm "${_tmp}"
+	return "${_err}"
 }
 
-# step_exec [-X] -b build-dir -l log -s step
+# step_exec [-X] -l log -s step
 #
 # Execute the given script and redirect any output to log.
 step_exec() (
-	local _builddir
 	local _err
 	local _fail
 	local _log
 	local _step
 	local _trace="yes"
+	local _tmpdir
 
 	while [ $# -gt 0 ]; do
 		case "$1" in
 		-X)	_trace="";;
-		-b)	shift; _builddir="$1";;
 		-l)	shift; _log="$1";;
 		-s)	shift; _step="$1";;
 		*)	break;;
 		esac
 		shift
 	done
-	: "${_builddir:?}"
 	: "${_log:?}"
 	: "${_step:?}"
 
-	_fail="$(mktemp -p "${_builddir}/tmp" step-exec.XXXXXX)"
-	echo 0 >"$_fail"
+	_tmpdir="$(config_value tmp-dir)"
+	_fail="$(mktemp -p "${_tmpdir}" step-exec.XXXXXX)"
+	echo 0 >"${_fail}"
 
-	[ "$DETACH" -eq 0 ] || exec >/dev/null 2>&1
+	[ "${DETACH}" -eq 0 ] || exec >/dev/null 2>&1
 
 	{
-		"$ROBSDEXEC" -m "$_MODE" ${ROBSDCONF:+"-C${ROBSDCONF}"} \
-			${_trace:+-x} "$_step" || echo "$?" >"$_fail"
+		"${ROBSDEXEC}" -m "${_MODE}" ${ROBSDCONF:+"-C${ROBSDCONF}"} \
+			${_trace:+-x} "${_step}" || echo "$?" >"${_fail}"
 
-		if [ "$_MODE" = "robsd-regress" ]; then
+		if [ "${_MODE}" = "robsd-regress" ]; then
 			# Regress tests can fail but still exit zero, check the
 			# log for failures.
-			regress_failed "$_log" && echo 1 >"$_fail"
+			regress_failed "${_log}" && echo 1 >"${_fail}"
 		fi
-	} </dev/null 2>&1 | tee "$_log"
-	_err="$(<"$_fail")"
-	rm -f "$_fail"
-	return "$_err"
+	} </dev/null 2>&1 | tee "${_log}"
+	_err="$(<"${_fail}")"
+	rm -f "${_fail}"
+	return "${_err}"
 )
 
 # step_exec_job -b build-dir -s steps -i step-id -n step-name
@@ -1139,55 +1194,33 @@ step_exec_job() {
 	: "${_name:?}"
 	: "${_steps:?}"
 
-	_log="$(log_id -b "$_builddir" -n "$_name" -s "$_id")"
+	_log="$(log_id -b "${_builddir}" -n "${_name}" -s "${_id}")"
 	_t0="$(date '+%s')"
-	step_write -t -l "$_log" -s "$_id" -n "$_name" -e -1 -d -1 "$_steps"
-	step_exec -b "$_builddir" -l "${_builddir}/${_log}" -s "$_name" || _exit="$?"
+	step_write -t -l "${_log}" -s "${_id}" -n "${_name}" -e -1 -d -1 "${_steps}"
+	step_exec -l "${_builddir}/${_log}" -s "${_name}" || _exit="$?"
 	_t1="$(date '+%s')"
 	_d1="$((_t1 - _t0))"
-	_d0="$(duration_prev "$_name" || :)"
-	if [ -n "$_d0" ]; then
+	_d0="$(duration_prev "${_name}" || :)"
+	if [ -n "${_d0}" ]; then
 		_delta="$((_d1 - _d0))"
 	else
 		_delta=0
 	fi
-	step_write -l "$_log" -s "$_id" -n "$_name" -e "$_exit" -d "$_d1" \
-		-a "$_delta" "$_steps"
+	step_write -l "${_log}" -s "${_id}" -n "${_name}" -e "${_exit}" -d "${_d1}" \
+		-a "${_delta}" "${_steps}"
 
-	robsd_hook -v "exit=${_exit}" -v "step=${_name}"
+	robsd_hook -v "step-exit=${_exit}" -v "step-name=${_name}"
 
-	case "$_MODE" in
+	case "${_MODE}" in
 	robsd-regress)
-		regress_step_after -b "$_builddir" -e "$_exit" -n "$_name" || return 1
+		regress_step_after -b "${_builddir}" -e "${_exit}" -n "${_name}" || return 1
 		;;
 	*)
-		[ "$_exit" -eq 0 ] || return 1
+		[ "${_exit}" -eq 0 ] || return 1
 		;;
 	esac
 }
 
-# step_field step-name
-#
-# Get the corresponding _STEP array index for the given field name.
-step_field() {
-	local _name
-
-	_name="$1"; : "${_name:?}"
-
-	case "$_name" in
-	step)		echo 0;;
-	name)		echo 1;;
-	exit)		echo 2;;
-	duration)	echo 3;;
-	log)		echo 4;;
-	time)		echo 5;;
-	user)		echo 6;;
-	skip)		echo 7;;
-	delta)		echo 8;;
-	*)		echo -1;;
-	esac
-}
-
 # step_id step-name
 #
 # Resolve the given step name to its corresponding numeric id.
@@ -1197,30 +1230,16 @@ step_id() {
 
 	_name="$1"; : "${_name:?}"
 
-	_id="$(steps | cat -n | grep -w "$_name" | awk '{print $1}')"
-	if [ -n "$_id" ]; then
-		echo "$_id"
+	_id="$(steps | grep -w "${_name}" | cut -d ' ' -f 1)"
+	if [ -n "${_id}" ]; then
+		echo "${_id}"
 	else
 		echo "step_id: ${_name}: unknown step" 1>&2
 		return 1
 	fi
 }
 
-# step_name step-id
-#
-# Resolve the given numeric step to its corresponding name.
-step_name() {
-	local _step
-
-	_step="$(steps | sed -n -e "${1}p")"
-	if [ -n "$_step" ]; then
-		echo "$_step"
-	else
-		return 1
-	fi
-}
-
-# step_path dir
+# step_path build-dir
 #
 # Get the path to the steps file.
 step_path() {
@@ -1230,11 +1249,11 @@ step_path() {
 	echo "${_dir}/step.csv"
 }
 
-# steps
+# steps [robsd-step-argument ...]
 #
-# Get the names of all steps in execution order.
+# Get steps in execution order.
 steps() {
-	"$ROBSDSTEP" -L -m "$_MODE" ${ROBSDCONF:+"-C${ROBSDCONF}"}
+	"${ROBSDSTEP}" -L -m "${_MODE}" ${ROBSDCONF:+"-C${ROBSDCONF}"} "$@"
 }
 
 # step_next file
@@ -1250,7 +1269,7 @@ step_next() {
 
 	_file="$1"; : "${_file:?}"
 
-	while step_eval "-${_i}" "$_file"; do
+	while step_eval "-${_i}" "${_file}"; do
 		_i="$((_i + 1))"
 
 		# The skip field is optional, suppress errors.
@@ -1260,10 +1279,10 @@ step_next() {
 
 		_step="$(step_value step)"
 		_exit="$(step_value exit)"
-		if [ "$_exit" -ne 0 ]; then
-			echo "$_step"
+		if [ "${_exit}" -ne 0 ]; then
+			echo "${_step}"
 		elif [ "$(step_value name)" = "end" ]; then
-			echo "$_step"
+			echo "${_step}"
 		else
 			echo $((_step + 1))
 		fi
@@ -1274,20 +1293,6 @@ step_next() {
 	return 1
 }
 
-# step_parallel step-name
-#
-# Exits zero if the step can be executed in parallel.
-step_parallel() {
-	case "$_MODE" in
-	robsd-regress)
-		regress_step_parallel "$_name"
-		;;
-	*)
-		return 1
-		;;
-	esac
-}
-
 # step_skip
 #
 # Exits zero if the step has been skipped.
@@ -1296,24 +1301,21 @@ step_skip() {
 
 	# The skip field is optional, suppress errors.
 	_skip="$(step_value skip 2>/dev/null)"
-	[ "$_skip" -eq 1 ]
+	[ "${_skip}" -eq 1 ]
 }
 
 # step_value field-name
 #
-# Get corresponding value for the given field name in the global _STEP array.
+# Get corresponding value for the given step field name.
 step_value() {
 	local _name
-	local _i
 
 	_name="$1"; : "${_name:?}"
 
-	_i="$(step_field "$_name")"
-	if [ "$_i" -lt 0 ] || [ -z "${_STEP[$_i]:-}" ]; then
+	if ! (eval "echo \${_STEP_${_name}}") 2>/dev/null; then
 		echo "step_value: ${_name}: unknown field" 1>&2
 		return 1
 	fi
-	echo "${_STEP[$_i]}"
 }
 
 # trap_exit -r robsd-dir [-b build-dir] [-s stat-pid]
@@ -1322,6 +1324,7 @@ step_value() {
 trap_exit() {
 	local _err="$?"
 	local _builddir=""
+	local _receiver=
 	local _robsddir=""
 	local _statpid=""
 	local _steps
@@ -1339,33 +1342,34 @@ trap_exit() {
 	done
 	: "${_robsddir:?}"
 
-	[ -z "$_statpid" ] || kill "$_statpid" || :
+	[ -z "${_statpid}" ] || kill "${_statpid}" || :
 
-	[ -n "$_builddir" ] || return "$_err"
+	[ -n "${_builddir}" ] || return "${_err}"
 
-	_steps="$(step_path "$_builddir")"
+	_steps="$(step_path "${_builddir}")"
 
 	# Generate the report if a step failed or the end step is reached.
-	if [ "$_err" -ne 0 ] ||
-	   step_eval -n end "$_steps" 2>/dev/null
+	if has_steps "${_steps}" &&
+	   { [ "${_err}" -ne 0 ] || step_eval -n end "${_steps}" 2>/dev/null; }
 	then
-		if report -b "$_builddir" &&
-		   [ "$DETACH" -ne 0 ]; then
-			# Do not send mail during interactive invocations.
-			sendmail root <"${_builddir}/report"
+		# Do not send mail during interactive invocations.
+		if report -b "${_builddir}" &&
+		   [ "${DETACH}" -ne 0 ]; then
+			_receiver="$(report_receiver -b "${_builddir}")"
+			sendmail "${_receiver}" <"$(config_value report-path)"
 		fi
 	fi
 
-	if step_eval -n end "$_steps" 2>/dev/null; then
-		robsd_hook -v "exit=0" -v "step=end"
+	if step_eval -n end "${_steps}" 2>/dev/null; then
+		robsd_hook -v "step-exit=0" -v "step-name=end"
 	fi
 
-	lock_release "$_robsddir" "$_builddir" || :
+	lock_release "${_robsddir}" "${_builddir}" || :
 
 	# Do not leave an empty build around.
-	[ -s "$_steps" ] || rm -r "$_builddir"
+	has_steps "${_steps}" || rm -r "${_builddir}"
 
-	return "$_err"
+	return "${_err}"
 }
 
 # unpriv [-c class] user utility argument ...
@@ -1389,11 +1393,11 @@ unpriv() (
 	# Since robsd is running as root, su(1) will preserve the following
 	# environment variables which is unwanted by especially some regress
 	# tests.
-	LOGNAME="$_user"
-	USER="$_user"
+	LOGNAME="${_user}"
+	USER="${_user}"
 	if [ $# -gt 0 ]; then
-		su ${_class:+-c "$_class"} "$_user" -c "$@"
+		su ${_class:+-c "${_class}"} "${_user}" -c "$@"
 	else
-		su ${_class:+-c "$_class"} "$_user" -sx
+		su ${_class:+-c "${_class}"} "${_user}" -sx
 	fi
 )
